/*
 * if_ext_act.c - Action database handling for if_ext
 *
 * Written 2001 by Werner Almesberger
 * Copyright 2001 by Network Robots
 */


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "config.h"
#include "util.h"
#include "error.h"
#include "data.h"
#include "op.h"
#include "iflib.h"


/*
 * Right now, there are two types of lists of actions, doing exactly the
 * same thing. Might as well merge them ...
 */

enum action_type { at_conform,at_count,at_class,at_drop };

static struct action {
    enum action_type type;
    union {
	int conform;		/* number of bucket */
	int count;		/* number of buckt */
	uint32_t decision;	/* qdisc:class */
    } u;
    struct action *parent;	/* parent in tree being constructed */
    struct action *c[2];	/* child if 0/1 */
    int number;			/* action number; -1 if none assigned */
    int dumped;			/* non-zero if entry has been dumped */
    struct action *next;	/* next action in list of entry points */
    struct referer *referers;	/* referers in inverted tree */
} *root = NULL;

struct referer {
    struct action *action;
    struct referer *next;
};


/* ---------------------- Step 1: Build the bit tree ----------------------- */


static struct action *make_action(DATA d,struct action *parent)
{
    struct action *action;

    action = alloc_t(struct action);
    if (d.op && d.op->dsc == &op_conform) {
	action->type = at_conform;
	action->u.conform = (int) d.op->a.u.police->number;
    }
    else if (d.op && d.op->dsc == &op_count) {
	action->type = at_count;
	action->u.count = (int) d.op->a.u.police->number;
    }
    else if (!d.op && d.type == dt_decision) {
	switch (d.u.decision.result) {
	    case dr_class:
		action->type = at_class;
		action->u.decision =
		  (d.u.decision.class->parent.qdisc->number << 16) |
		  d.u.decision.class->number;
		break;
	    case dr_drop:
		action->type = at_drop;
		break;
	    default:
		dump_failed(d);
	}
    }
    else abort();
    action->parent = parent;
    action->c[0] = action->c[1] = NULL;
    action->number = -1;
    action->dumped = 0;
    action->next = NULL;
    action->referers = NULL;
    return action;
}


static struct action *add_actions(DATA d,DATA next,DATA otherwise,
  struct action *parent)
{
    struct action *action;

    if (debug > 1) debugf("add_actions");
    if (d.op && d.op->dsc == &op_logical_or) {
	assert(!next.op && next.type == dt_none);
	assert(!otherwise.op && otherwise.type == dt_none);
	return add_actions(d.op->a,data_none(),d.op->b,parent);
    }
    if (d.op && d.op->dsc == &op_logical_and) {
	assert(!next.op && next.type == dt_none); /* solution ? @@@ */
	return add_actions(d.op->a,d.op->b,otherwise,parent);
    }
    if (d.op && d.op->dsc == &op_conform) {
	action = make_action(d,parent);
	action->c[1] = add_actions(next,data_none(),otherwise,action);
	action->c[0] = add_actions(otherwise,data_none(),data_none(),action);
	return action;
    }
    if (d.op && d.op->dsc == &op_count) {
	action = make_action(d,parent);
	action->c[1] = add_actions(next,data_none(),otherwise,action);
	return action;
    }
    if (!d.op && d.type == dt_decision)
	return make_action(d,parent);
    dump_failed(d);
    return NULL; /* not reached */
}


/* --------- Step 2: Eliminate nodes repeating previous decisions ---------- */


static void free_1(struct action *a)
{
    free(a);
}


static void free_subtree(struct action *a)
{
    if (a) return;
    free_subtree(a->c[0]);
    free_subtree(a->c[1]);
    free_1(a);
}


static void repeated_actions(struct action *a)
{
    if (!a) return;
    if (a->type == at_conform) {
	struct action *walk;

	for (walk = a; walk; walk = walk->parent) {
	    struct action **anchor;
	    int value;

	    if (!walk->parent || walk->parent->type != at_conform ||
	      walk->parent->u.conform != a->u.conform)
		continue;
	    value = walk->parent->c[0] == walk ? 0 : 1;
	    anchor = a->parent->c[0] == a ? &a->parent->c[0] : &a->parent->c[1];
	    *anchor = a->c[value];
	    a->c[value]->parent = a->parent;
	    free_subtree(a->c[1-value]);
	    free_1(a);
	    repeated_actions(*anchor);
	    return;
	}
    }
    repeated_actions(a->c[0]);
    repeated_actions(a->c[1]);
}


/* ----------------------- Hash table for leaf nodes ----------------------- */


#define HASH_ENTRIES 47


static struct hash_list {
    struct action *action;
    struct hash_list *next;
} *hash[HASH_ENTRIES];


static struct action *hash_find(struct action *a)
{
    struct hash_list *walk,*new;
    int key;

    assert(a->type == at_class || a->type == at_drop);
    key = a->u.decision % HASH_ENTRIES;
    for (walk = hash[key]; walk; walk = walk->next)
	if (walk->action->u.decision == a->u.decision)
	    return walk->action;
    new = alloc_t(struct hash_list);
    new->action = a;
    new->next = hash[key];
    hash[key] = new;
    return a;
}


/* --------------------- Step 3: Merge common subtrees --------------------- */


static int action_eq(const struct action *a,const struct action *b)
{
    if (a->type != b->type) return 0;
    switch (a->type) {
	case at_class:
	    return a->u.decision == b->u.decision;
	case at_drop:
	    return 1;
	case at_conform:
	    return a->u.conform == b->u.conform;
	case at_count:
	    return a->u.count == b->u.count;
	default:
	    abort();
    }
}


static void add_referer(struct action *action,struct action *referer)
{
    struct referer **here;

    for (here = &action->referers; *here; here = &(*here)->next);
    *here = alloc_t(struct referer);
    (*here)->action = referer;
    (*here)->next = NULL;
}


static struct action *unique_action(struct action *a);

static struct action *_unique_action(struct action *a)
{
    struct referer *referer,*walk;

    switch (a->type) {
	case at_class:
	case at_drop:
	    return hash_find(a);
	case at_conform:
	    a->c[0] = unique_action(a->c[0]);
	    a->c[1] = unique_action(a->c[1]);
	    for (referer = a->c[0]->referers; referer; referer = referer->next)
		if (action_eq(referer->action,a))
		    for (walk = a->c[1]->referers; walk; walk = walk->next)
			if (referer->action == walk->action)
			    return referer->action;
	    add_referer(a->c[0],a);
	    add_referer(a->c[1],a);
	    return a;
	case at_count:
	    a->c[1] = unique_action(a->c[1]);
	    for (referer = a->c[1]->referers; referer; referer = referer->next)
		if (action_eq(referer->action,a)) return referer->action;
	    add_referer(a->c[1],a);
	    return a;
	default:
	    abort();
    }
}


static struct action *unique_action(struct action *a)
{
    struct action *w;

    if (debug > 2) {
	for (w = a; w; w = w->parent) fprintf(stderr,"  ");
	fprintf(stderr,"ENTER unique_action: %p\n",a);
    }
    a = _unique_action(a);
    if (debug > 2) {
	for (w = a; w; w = w->parent) fprintf(stderr,"  ");
	fprintf(stderr,"EXIT unique_action: %p\n",a);
    }
    return a;
}


static void to_unique(struct action *a)
{
    struct action *res,*walk;

    res = unique_action(a);
    for (walk = root; walk; walk = walk->next)
	if (res == walk) return;
    res->next = root;
    root = res;
}


/* ---------------------------- Dump a subtree ----------------------------- */


static void dump_items(FILE *file,const struct action *a,int top)
{
    if (a->number != -1 && !top) {
	fprintf(file," action %d",a->number);
	return;
    }
    switch (a->type) {
	case at_conform:
	    fprintf(file," conform %d",a->u.conform);
	    dump_items(file,a->c[1],0);
	    dump_items(file,a->c[0],0);
	    break;
	case at_count:
	    fprintf(file," count %d",a->u.count);
	    dump_items(file,a->c[1],0);
	    break;
	case at_class:
	    fprintf(file," class %u:%u",a->u.decision >> 16,
	      a->u.decision & 0xffff);
	    break;
	case at_drop:
	    fprintf(file," drop");
	    break;
	default:
	    abort();
    }
}


static void dump_subtree(FILE *file,struct action *a,int top)
{
    switch (a->type) {
	case at_conform:
	    dump_subtree(file,a->c[0],0);
	    /* fall through */
	case at_count:
	    dump_subtree(file,a->c[1],0);
	    /* fall through */
	case at_class:
	case at_drop:
	    break;
	default:
	    abort();
    }
    if ((top || a->referers->next) && !a->dumped) {
	a->dumped = 1;
	fprintf(file,"action %d =",a->number);
	dump_items(file,a,1);
	fputc('\n',file);
    }
}


static int number_taken(const struct action *a,int index)
{
    if (!a) return 0;
    if (a->number == index) return 1;
    return number_taken(a->c[0],index) || number_taken(a->c[1],index);
}


static int number_available(const struct action *root,int index)
{
    const struct action *walk;

    for (walk = root; walk; walk = walk->next)
	if (number_taken(walk,index)) return 0;
    return 1;
}


static void assign_indices(const struct action *root,struct action *a,int top,
  int preferred,int *index)
{
    switch (a->type) {
	case at_conform:
	    assign_indices(root,a->c[0],0,preferred,index);
	    /* fall through */
	case at_count:
	    assign_indices(root,a->c[1],0,preferred,index);
	    /* fall through */
	case at_drop:
	    break;
	case at_class:
	    if ((top || a->referers->next) && a->number == -1 && preferred &&
	      number_available(root,a->u.decision & 0xff))
		a->number = a->u.decision & 0xff;
	    break;
	default:
	    abort();
    }
    if (!(top || a->referers->next) || a->number != -1 || preferred) return;
    while (!number_available(root,*index)) (*index)++;
    a->number = (*index)++;
}


/* -------------------------- Debugging tree dump -------------------------- */


static void debug_subtree(FILE *file,const struct action *a,int level)
{
    const struct referer *r;

    fprintf(file,"%*s[%p] %d ",level*4,"",a,a->number);
    switch (a->type) {
	case at_conform:
	    fprintf(file,"conform %d (0, then 1)\n",a->u.conform);
	    break;
	case at_count:
	    fprintf(file,"count %d\n",a->u.count);
	    break;
	case at_class:
	    fprintf(file,"class %u:%u\n",a->u.decision >> 16,
	      a->u.decision & 0xffff);
	    break;
	case at_drop:
	    fprintf(file,"drop\n");
	    break;
	default:
	    abort();
    }
    fprintf(file,"%*s| parent %p, next %p\n",level*4,"",a->parent,a->next);
    fprintf(file,"%*s| referers:",level*4,"");
    for (r = a->referers; r; r = r->next)
	fprintf(file," %p",r->action);
    fputc('\n',file);
    switch (a->type) {
	case at_conform:
	    debug_subtree(file,a->c[0],level+1);
	    /* fall through */
	case at_count:
	    debug_subtree(file,a->c[1],level+1);
	    /* fall through */
	case at_class:
	case at_drop:
	    break;
	default:
	    abort();
    }
}


static void debug_action_tree(FILE *file)
{
    struct action *walk;

    for (walk = root; walk; walk = walk->next)
	debug_subtree(file,walk,0);
}


/* ------------------------------------------------------------------------- */


void register_actions(DATA d)
{
    struct action *action;

    action = add_actions(d,data_none(),data_none(),NULL);
    repeated_actions(action);
    to_unique(action);
}


void write_actions(FILE *file)
{
    static int index = 0;
    struct action *walk;

    /* assign preferred numbers */
    for (walk = root; walk; walk = walk->next)
	assign_indices(root,walk,1,1,&index);
    /* assign all other numbers */
    for (walk = root; walk; walk = walk->next)
	assign_indices(root,walk,1,0,&index);
    /* dump the thing */
    for (walk = root; walk; walk = walk->next)
	dump_subtree(file,walk,1);
}


int action_number(DATA d)
{
    struct action *action;

    if (debug > 1) debug_action_tree(stderr);
    action = add_actions(d,data_none(),data_none(),NULL);
    repeated_actions(action);
    if (debug > 1) fprintf(stderr,"in-action %p\n",action);
    action = unique_action(action);
    if (debug > 1) {
	fprintf(stderr,"out-action %p\n",action);
	print_expr(stderr,d);
    }
    return action->number;
}


void free_actions(void)
{
    /* tricky: walk tree while == 1 referer */
}
